#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/wait.h>

#define and &&
#define or ||
#define FATAL_ERROR(X)\
	{printf("%s::%lu:%s\n", __FILE__, (unsigned long) __LINE__, X); exit(EXIT_FAILURE);}
#define BUG(X) {if (X) FATAL_ERROR(#X)}
#define ABS(X) ((X) > 0 ? (X) : (-(X)))

#define NPV_MAX 256
#define PVSTRING_MAX 16

typedef struct
{
	char pv_string[PVSTRING_MAX];
	long pv_move;
	int pv_score;
} pv_t;

#define HOSTS_CFG "hosts.cfg"
#define HOSTS_MAX 256
#define HOSTNAME_MAX 32
#define USERNAME_MAX 32
#define PATHNAME_MAX 64
#ifndef LINE_MAX
#define LINE_MAX 255
#endif
#ifndef TRUE
#define TRUE 1
#endif
#undef NFS

typedef struct
{
	char host_name[HOSTNAME_MAX];
	char host_id[HOSTNAME_MAX];
	char user_name[USERNAME_MAX];
	char host_path[PATHNAME_MAX];
	char host_output[PATHNAME_MAX];
	char host_dbase[PATHNAME_MAX];
	int host_npv;
	pv_t host_pv[NPV_MAX];
	pid_t host_pid;
} host_t;

int npv;
pv_t pv[NPV_MAX];

int nhosts;
host_t hosts[HOSTS_MAX];

#define STRING_MAX 16
#define TOURNAMENT
#ifndef TOURNAMENT
#define GAME_TIME1 (120)
#define GAME_TIME2 (30)
#define NGAME_MOVES1 10
#define NGAME_MOVES2 5
#else
#define GAME_TIME1 (90 * 60)
#define GAME_TIME2 (60 * 60)
#define NGAME_MOVES1 60
#define NGAME_MOVES2 40
#define GAME_MARGIN 5
#endif

char ZColour[STRING_MAX];
int GameTimeUsed = 0;
int TimeLimit = 75;
char colour_to_move[STRING_MAX];
int ngame_moves = 0;
time_t wall1, wall2;
int wall;

#define PROGRAM    "zzzzzz"
#define DBASE_FILE "dbase.txt"
#define GAME_FILE  "game.txt"

int main(int, char **);
int main(int argc, char **argv)
{
	int err;
	char line[LINE_MAX], move[LINE_MAX];
	FILE *f;
	int ihost;
	int ipv, jpv, npv;

	argc--;
	argv++;
	BUG(argc != 2)
	BUG(sscanf(argv[0], "%d", &GameTimeUsed) != 1)
	GameTimeUsed *= 60;
	BUG(sscanf(argv[1], "%s", ZColour) != 1)
	BUG((strcmp(ZColour, "white") != 0) and (strcmp(ZColour, "black") != 0))
	printf("GameTimeUsed=%d Zcolour=%s\n", GameTimeUsed, ZColour);

	/* read hosts database */

	BUG((f = fopen(HOSTS_CFG, "r")) == NULL)
	nhosts = 0;
	while(TRUE)
	{
		host_t *with;

		BUG(nhosts >= HOSTS_MAX)
		with = hosts + nhosts;
		if (fscanf(f, "%s%s%s%s",
			with->host_name, with->host_id,
			with->user_name, with->host_path) != 4) break;
		sprintf(with->host_dbase, "dbase.%s", with->host_id);
		sprintf(with->host_output, "out.%s", with->host_id);
		nhosts++;
	}
	BUG(fclose(f) == EOF)

	label_next_move:

	wall1 = time(NULL);
	sprintf(line, PROGRAM " what " GAME_FILE " " DBASE_FILE);
	system(line);
	BUG((f = fopen(DBASE_FILE, "r")) == NULL)
	BUG(fgets(line, LINE_MAX, f) == NULL)
	BUG(sscanf(line, "%s%d", colour_to_move, &ngame_moves) != 2)
	BUG(fgets(line, LINE_MAX, f) == NULL)
	BUG(sscanf(line, "%d", &npv) != 1)
	if (npv == 0)
	{
		BUG(fgets(line, LINE_MAX, f) == NULL)
		printf("%s %s", colour_to_move, line);	
		exit(EXIT_FAILURE);
	}
	for (ipv = 0; ipv < ABS(npv); ipv++)
	{
		BUG(fgets(line, LINE_MAX, f) == NULL)
		BUG(sscanf(line, "%s%lx", pv[ipv].pv_string, &(pv[ipv].pv_move)) != 2)
	}
	/*
	for (ipv = 0; ipv < ABS(npv); ipv++)
		printf("%s %08lx\n", pv[ipv].pv_string, pv[ipv].pv_move);
	*/
	if (strcmp(ZColour, colour_to_move) == 0) goto label_zzzzzz;

	if ((ngame_moves > 0) and (ngame_moves % 5) == 0)
	{
		while(TRUE)
		{
			printf("\aGameTimeUsed? (min) ");
			if (gets(line) == NULL) exit(EXIT_FAILURE);
			if (sscanf(line, "%d", &GameTimeUsed) != 1) continue;
			GameTimeUsed *= 60;
			break;
		}
	}

	while(TRUE)
	{
		printf("your move? ");
		if (gets(line) == NULL) exit(EXIT_FAILURE);
		if (*line == '!')
		{
			system(line + 1);
			continue;
		}
		if (sscanf(line, "%s", move) != 1) continue;
		if (strcmp(move, "quit") == 0) exit(EXIT_SUCCESS);
		if (strcmp(move, "?") == 0)
		{
			for (ipv = 0; ipv < ABS(npv); ipv++)
			{
				printf(" %5s", pv[ipv].pv_string);
				if (((ipv % 10) == 9) or (ipv == (npv - 1))) printf("\n");
			}
			continue;
		}
		for (ipv = 0; ipv < ABS(npv); ipv++)
			if (strcmp(move, pv[ipv].pv_string) == 0) break;
		if (ipv == ABS(npv)) continue;
		BUG((f = fopen(GAME_FILE, "a")) == NULL)
		fprintf(f, "%s\n", pv[ipv].pv_string);
		BUG(fclose(f) == EOF)
		break;
	}
	goto label_next_move;

	label_zzzzzz:
	if (npv == 1)
	{
		printf("forced!\n");
		jpv = 0;
		goto no_problem;
	}
	if (npv < 0)
	{
		printf("book move!\n");
		jpv = 0;
		goto no_problem;
	}
	BUG(npv < 0)

	/* calculate time limit */

	if (ngame_moves < NGAME_MOVES1)
	{
		TimeLimit =
		(GAME_TIME1 - GAME_MARGIN - GameTimeUsed) /
	 	(NGAME_MOVES1 - ngame_moves);
	}
	else if (ngame_moves < (NGAME_MOVES1 + NGAME_MOVES2))
	{
		TimeLimit =
		(GAME_TIME1 + GAME_TIME2 - GAME_MARGIN - GameTimeUsed) / 
		(NGAME_MOVES1 + NGAME_MOVES2 - ngame_moves);
	}
	else
	{
		TimeLimit = 90;
	}
	if (TimeLimit < 0)
	{
		printf("warning: negative time limit!\n");
		TimeLimit = 1;
	}
	printf("TimeLimit=%d\n", TimeLimit);

	
	/* distribute pv */
	
	for (ihost = 0; ihost < nhosts; ihost++)
		hosts[ihost].host_npv = 0;

	ihost = 0;
	for (ipv = 0; ipv < npv; ipv++)
	{
		host_t *with;

		if (ihost == nhosts) ihost = 0;
		with = hosts + ihost;
		with->host_pv[with->host_npv] = pv[ipv];
		with->host_npv++;
		ihost++;
	}
	ipv = 0;
	for (ihost = 0; ihost < nhosts; ihost++)
		ipv += hosts[ihost].host_npv;
	BUG(ipv != npv);

	for (ihost = 0; ihost < nhosts; ihost++)
	{
		host_t *with = hosts + ihost;
		pid_t pid;

		if (with->host_npv == 0) continue;
		BUG((pid = fork()) < 0)
		if (pid == 0)
		{
			char line[80];
			FILE *f;

			BUG((f = fopen(with->host_dbase, "w")) == NULL)
			fprintf(f, "%d\n", with->host_npv);
			for (ipv = 0; ipv < with->host_npv; ipv++)
			{
				fprintf(f, "%s %08lx\n",
					with->host_pv[ipv].pv_string, with->host_pv[ipv].pv_move);
			}
			BUG(fclose(f) == EOF)
#ifdef NFS
			sprintf(line, "cp %s %s", with->host_dbase, with->host_path);
			err = system(line);
			if (err != 0)
				printf("cp dbase failure for %s!\n", with->host_id);
			sprintf(line, "cp " GAME_FILE " %s", with->host_path);
			err = system(line);
			if (err != 0)
				printf("cp game failure for %s!\n", with->host_id);
#else
			sprintf(line, "rcp %s %s@%s:%s",
				with->host_dbase, with->user_name, with->host_name, with->host_path);
			err = system(line);
			if (err != 0)
				printf("rcp dbase failure for %s!\n", with->host_id);
			sprintf(line, "rcp " GAME_FILE " %s@%s:%s", with->user_name,
				with->host_name, with->host_path);
			err = system(line);
			if (err != 0)
				printf("rcp game failure for %s!\n", with->host_id);
#endif
			sprintf(line, "remsh %s -l %s 'cd %s; " PROGRAM 
				" -NoBook -DepthMax 32 -TimeLimit %d par " GAME_FILE " %s' >%s",
				with->host_name, with->user_name, with->host_path, 
				TimeLimit, with->host_dbase, with->host_output);
			/* printf("%s\n", line); */
			err = system(line);
			if (err != 0)
				printf("remsh failure for %s!\n", with->host_id);
#ifdef NFS
			sprintf(line, "cp %s%s ./", 
				with->host_path, with->host_dbase);
			err = system(line);
			if (err != 0)
				printf("cp failure for %s!\n", with->host_id);
#else
			sprintf(line, "rcp %s@%s:%s%s %s", 
				with->user_name, with->host_name,
			 	with->host_path, with->host_dbase, with->host_dbase);
			err = system(line);
			if (err != 0)
				printf("rcp failure for %s!\n", with->host_id);
#endif
			exit(0);
		}
		else
		{
			with->host_pid = pid;
			printf("child for %s pid=%d\n", with->host_id, with->host_pid);
		}
	}
	
	while(TRUE)
	{
		int ihost;

		for (ihost = 0; ihost < nhosts; ihost++)
		{
			if (hosts[ihost].host_npv == 0) continue;
			if (hosts[ihost].host_pid > 0) break;
		}
		if (ihost == nhosts) break;
		err = wait(NULL);
		printf("wait pid=%d\n", err);
		if (err < 0)
			printf("wait failed!\n");
		else
		{
			for (ihost = 0; ihost < nhosts; ihost++)
			{
				if (hosts[ihost].host_npv == 0) continue;
				if (hosts[ihost].host_pid == err) break;
			}
			if (ihost == nhosts)
				printf("wait returned invalid pid!\n");
			else
			{
				wall2 = time(NULL);
				wall = (int) (wall2 - wall1) + 1;
				printf("ready %s wall=%d\n", hosts[ihost].host_id, wall);
				hosts[ihost].host_pid = 0;
			}
		}
	}

	npv = 0;
	for (ihost = 0; ihost < nhosts; ihost++)
	{
		host_t *with = hosts + ihost;

		if (with->host_npv == 0) continue;
		BUG((f = fopen(with->host_dbase, "r")) == NULL)
		BUG(fscanf(f, "%s %d\n", pv[npv].pv_string, &(pv[npv].pv_score)) != 2);
		printf("%s %d\n", pv[npv].pv_string, pv[npv].pv_score);
		BUG(fclose(f) == EOF)
		npv++;
	}
	jpv = 0;
	for (ipv = 1; ipv < npv; ipv++)
	{
		if (pv[ipv].pv_score > pv[jpv].pv_score) jpv = ipv;
	}

	wall2 = time(NULL);
	wall = (int) (wall2 - wall1) + 1;
	GameTimeUsed += wall;
	printf("WallTime=%d GameTimeUsed=%d\n", wall, GameTimeUsed);

	no_problem:
	BUG((f = fopen(GAME_FILE, "a")) == NULL)
	fprintf(f, "%s zzzzzz %d\n", pv[jpv].pv_string, pv[jpv].pv_score);
	BUG(fclose(f) == EOF)
	printf("\azzzzzz move: %s %d\n", pv[jpv].pv_string, pv[jpv].pv_score);
	goto label_next_move;
}
